/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "rumd_data.h"

void CalcR2(Conf *C0, Conf *Ct, double *R2,  double *R4, double *Fs, double *Fs2, unsigned int num_types, double *q) {
  unsigned int num_part = C0->num_part;
  float L = C0->metaData.L[0];

  double *Fs_;
  Fs_ = new double[num_types];

  for (unsigned int i=0; i<num_types; i++) Fs_[i] = 0.0;

  for (unsigned int i=0; i<num_part; i++) {
    unsigned int Type = C0->P[i].MyType;
    assert(Type<num_types);
    assert(Type==Ct->P[i].MyType);
    double Rsq = C0->P[i].Rsq(&Ct->P[i], L, L, L );
    R2[Type]  += Rsq;
    R4[Type]  += Rsq*Rsq;
    Fs_[Type] += C0->P[i].Fsq(&Ct->P[i], L, L, L, q[Type]);
  }
  for (unsigned int i=0; i<num_types; i++) {
    Fs[i] += Fs_[i];
    Fs2[i] += Fs_[i]*Fs_[i];
  }
  delete[] Fs_;
}

void CalcR2cm(Conf *C0, Conf *Ct, double *R2,  double *R4, double *Fs, unsigned int num_types, double *q, unsigned int particlesPerMol) {
  // Calculate dynamical properties of center of mass of molecules, assuming they all have 'particlesPerMol' atoms

  unsigned int num_part = C0->num_part;
  double Lx = C0->metaData.L[0];
  double Ly = C0->metaData.L[1];
  double Lz = C0->metaData.L[2];

  for (unsigned int i=0; i<num_part; i+=particlesPerMol) {
    double dXtot   = 0.0;
    double dYtot   = 0.0;
    double dZtot   = 0.0;
    double totMass = 0.0;
   
    for (unsigned int j=0; j<particlesPerMol; j++) {
      unsigned int myType = C0->P[i+j].MyType;
      assert(myType<num_types);
      assert(myType==Ct->P[i+j].MyType);
      double myMass = C0->metaData.massOfType[myType];
      totMass += myMass;
 
      // Calculate the displacement of the i+j'th particle
      double dx = Ct->P[i+j].x - C0->P[i+j].x + Lx*(Ct->P[i+j].Imx - C0->P[i+j].Imx);
      double dy = Ct->P[i+j].y - C0->P[i+j].y + Ly*(Ct->P[i+j].Imy - C0->P[i+j].Imy);
      double dz = Ct->P[i+j].z - C0->P[i+j].z + Lz*(Ct->P[i+j].Imz - C0->P[i+j].Imz);
           
      // Add to the center of mass discplacement
      dXtot += myMass*dx;
      dYtot += myMass*dy;
      dZtot += myMass*dz;
    }                         
    
    dXtot /= totMass;
    dYtot /= totMass;
    dZtot /= totMass;
    
    double Rsq = dXtot*dXtot + dYtot*dYtot + dZtot*dZtot;
                    
    // We are not (yet) differentiating between different types of molecules
    R2[0] += Rsq;
    R4[0] += Rsq*Rsq;
    Fs[0] += (cos(dXtot*q[0]) + cos(dYtot*q[0]) + cos(dZtot*q[0]) )/3.0;
  }
}


int main(int argc, char *argv[])
{
  unsigned int first_block=0;
  unsigned int last_block;   
  unsigned int particlesPerMol = 0;
  std::string directory("TrajectoryFiles");

  static const char *usage = "Usage: %s [-h] [-m <particles per molecule>] [-d directory]\n";
  int opt;
  while ((opt = getopt(argc, argv, "hm:d:")) != -1) {
    switch (opt) {
    case 'm':
      particlesPerMol = atoi(optarg);
      break;
    case 'h':
      fprintf(stdout, usage, argv[0]);
      exit(EXIT_SUCCESS);
    case 'd':
      directory = std::string(optarg);
      break;
    case '?':
      fprintf(stderr, "%s: unknown option: -%c\n", argv[0], optopt);
      fprintf(stderr, usage, argv[0]);
      exit(EXIT_FAILURE);
      break;
    case ':':
      fprintf(stderr, "%s: missing option argument for option: -%c\n", argv[0], optopt);
      fprintf(stderr, usage, argv[0]);
      exit(EXIT_FAILURE);
      break;
    default:
      fprintf(stderr, usage, argv[0]);
      exit(EXIT_FAILURE);
    }
  }

  // Change to more user friendly option reader:
  //sscanf(argv[1], "%d", &first_block);
  //sscanf(argv[2], "%d", &last_block);

  Conf *C0 = new Conf();
  Conf *Ct = new Conf();

  gzFile gzfp;
  FILE *fp;

  fp=fopen((directory + "/LastComplete_block.txt").c_str(),"r");
  if (!fp){
    fprintf (stderr, (std::string("Error opening file ") + directory + "/LastComplete_block.txt:" + strerror(errno) + "\n").c_str());
    exit(EXIT_FAILURE);
  }
  fscanf(fp, "%u", &last_block);
  fclose(fp);

  // Read metaData from first configuration in first block
  char *C0Name;
  asprintf(&C0Name, (directory + "/block%.4u.xyz.gz").c_str(), first_block);
  printf("Reading block %u: %s\n\n", first_block, C0Name);
  gzfp=gzopen(C0Name,"r");
  if (!gzfp){
    fprintf (stderr, "rumd_msd: Error opening file %s: %s\n",C0Name, strerror(errno));
    exit(EXIT_FAILURE);
  }
  C0->Read(gzfp,C0Name,1);
  gzclose(gzfp);
  free(C0Name);

  // Store the metaData we need
  // float L  = C0->metaData.L[0];     // Assuming Lx=Ly=Lz for now...
  float dt = C0->metaData.dt;
  unsigned int num_part = C0->num_part;
  unsigned int num_types = C0->metaData.numTypes;
  unsigned int MaxIndex = C0->metaData.logLin.maxIndex;

  double *qvalues;
  int *NoPartOfType;
  qvalues = new double[num_types];
  NoPartOfType = new int[num_types];

  fp=fopen("qvalues.dat","r");
  if (!fp){
    fprintf (stderr, "Error opening file qvalues.dat: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }
  printf("\nReading q-values: ");
  for (unsigned int i=0; i<num_types; i++) {
    fscanf(fp, "%lf", &qvalues[i]);
    printf("%f ", qvalues[i]);
    assert(qvalues[i]>0);
    NoPartOfType[i] = 0;
  }
  printf("\n\n");
  fclose(fp);

  // Read and print the number of particles of each type
  for (unsigned int i=0; i<num_part; i++) {
    unsigned int Type = C0->P[i].MyType;
    assert(Type<num_types);
    NoPartOfType[Type]++;
  }
  for (unsigned int i=0; i<num_types; i++) {
    printf("Particles of type %d: %d\n", i, NoPartOfType[i]);
  }

  unsigned int num_moleculetypes = 0;
  if (particlesPerMol>1) num_moleculetypes = 1; // For now only deal with one type of molecules at most 

  // Setup arrays to collect statistics
  unsigned long int *Time;
  unsigned long int *Count;
  double **R2;
  double **R4;
  double **Fs;
  double **Fs2;

  double **R2cm;
  double **R4cm;
  double **Fscm;


  int MaxDataIndex = MaxIndex + int(log2(last_block-first_block+1)+1);
  //printf("\nData structure info: %d %d %d %d\n", num_part, num_types, MaxIndex, MaxDataIndex);

  Time = new unsigned long int[MaxDataIndex];
  R2    = new double*[MaxDataIndex];
  R4    = new double*[MaxDataIndex];
  Fs    = new double*[MaxDataIndex];
  Fs2   = new double*[MaxDataIndex];
  R2cm  = new double*[MaxDataIndex];
  R4cm  = new double*[MaxDataIndex];
  Fscm  = new double*[MaxDataIndex];
  Count = new unsigned long int[MaxDataIndex];
  for (int i=0; i<MaxDataIndex; i++ ) {
    R2[i]    = new double[num_types];
    R4[i]    = new double[num_types];
    Fs[i]    = new double[num_types];
    Fs2[i]   = new double[num_types];
    R2cm[i]  = new double[num_moleculetypes];
    R4cm[i]  = new double[num_moleculetypes];
    Fscm[i]  = new double[num_moleculetypes];
    Count[i] = 0;
    for (unsigned int j=0; j<num_types; j++) {
      R2[i][j] = 0.0;
      R4[i][j] = 0.0;
      Fs[i][j] = 0.0;
      Fs2[i][j] = 0.0;
    }
    for (unsigned int j=0; j<num_moleculetypes; j++) {
      R2cm[i][j] = 0.0;
      R4cm[i][j] = 0.0;
      Fscm[i][j] = 0.0;
    }
  }

  printf("\nReading block: ");
  for (unsigned int i=first_block; i<=last_block; i++) {

    // Read first configuration of block 'i'
    char *C0Name;
    asprintf(&C0Name,(directory + "/block%.4u.xyz.gz").c_str(),i);
    if ((i-first_block)%10==0) printf("%u ", i);
    fflush(stdout);
    gzfp=gzopen(C0Name,"r");
    C0->Read(gzfp,C0Name,0);

    // Check that relevant metaData didn't change
    assert(C0->metaData.dt       == dt);
    assert(C0->num_part          == num_part);
    assert(C0->metaData.numTypes == num_types);
    assert(MaxIndex == C0->metaData.logLin.maxIndex);

    // Analyze configurations in the same block-file
    for (unsigned int j=1; j<=MaxIndex; j++) {
      Ct->Read(gzfp,C0Name,0);
      Count[j-1]++;
      CalcR2(C0, Ct, R2[j-1], R4[j-1], Fs[j-1], Fs2[j-1], num_types, qvalues);
      if (num_moleculetypes>0) {
        CalcR2cm(C0, Ct, R2cm[j-1], R4cm[j-1], Fscm[j-1], num_types, qvalues, particlesPerMol);
      }
    }
    gzclose(gzfp);
    free(C0Name);

    // Analyze configurations in the later block-files
    for (unsigned int j=1; i+(1<<j)<=last_block; j++) {
      char *C0Name;
      asprintf(&C0Name,(directory+"/block%.4u.xyz.gz").c_str(),i+(1<<j));
      gzfp=gzopen(C0Name,"r");
      Ct->Read(gzfp,C0Name,0);
      int jj = MaxIndex+j-1;
      assert(jj<MaxDataIndex);
      Count[jj]++;
      CalcR2(C0, Ct, R2[jj], R4[jj], Fs[jj], Fs2[jj], num_types, qvalues);
      if (num_moleculetypes>0) {
        CalcR2cm(C0, Ct, R2cm[jj], R4cm[jj], Fscm[jj], num_types, qvalues, particlesPerMol);
      }
      gzclose(gzfp);
      free(C0Name);
     }
  }
  printf("\n");

  fp = fopen("msd.dat","w");
  for (int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      fprintf(fp, "%f ", (1UL<<i)*dt);
      for (unsigned int j=0; j<num_types; j++) {
        fprintf(fp, "%f ", R2[i][j]/float(Count[i]*NoPartOfType[j]));
      }
      fprintf(fp, "\n ");
    }
  }
  fclose(fp);

  if (num_moleculetypes>0) {
    fp = fopen("msd_cm.dat","w");
    for (int i=0; i<MaxDataIndex; i++ ) {
      if (Count[i]>0) {
        fprintf(fp, "%f ", (1UL<<i)*dt);
        for (unsigned int j=0; j<num_moleculetypes; j++) {
	  fprintf(fp, "%f ", R2cm[i][j]/float(Count[i]*num_part)*float(particlesPerMol));
        }
        fprintf(fp, "\n ");
      }
    }
    fclose(fp);
  }

  fp = fopen("alpha2.dat","w");
  for (int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      fprintf(fp, "%f ", (1UL<<i)*dt);
      for (unsigned int j=0; j<num_types; j++) {
	fprintf(fp, "%f ", 3.f/5.f*R4[i][j]/powf(R2[i][j],2.f)*float(Count[i]*NoPartOfType[j])-1.f);
      }
      fprintf(fp, "\n ");
    }
  }

  fp = fopen("Fs.dat","w");
  for (int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      fprintf(fp, "%f ", (1UL<<i)*dt);
      for (unsigned int j=0; j<num_types; j++) {
	fprintf(fp, "%f ", Fs[i][j]/float(Count[i]*NoPartOfType[j]));
      }
      fprintf(fp, "\n ");
    }
  }
  fclose(fp);
 
  if (num_moleculetypes>0) {
    fp = fopen("Fs_cm.dat","w");
    for (int i=0; i<MaxDataIndex; i++ ) {
      if (Count[i]>0) {
        fprintf(fp, "%f ", (1UL<<i)*dt);
        for (unsigned int j=0; j<num_moleculetypes; j++) {
	  fprintf(fp, "%f ", Fscm[i][j]/float(Count[i]*num_part)*float(particlesPerMol));
        }
        fprintf(fp, "\n ");
      }
    }
    fclose(fp);
  }

  fp = fopen("FsSq.dat","w");
  for (int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      fprintf(fp, "%f ", (1UL<<i)*dt);
      for (unsigned int j=0; j<num_types; j++) {
	fprintf(fp, "%f ", Fs2[i][j]/float(Count[i]*NoPartOfType[j]) );
      }
      fprintf(fp, "\n ");
    }
  }
  fclose(fp);

}
