/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/
#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <cerrno>
#include <cmath>
#include <cassert>
#include <string>
#include <iostream>
#include <getopt.h>
#include "rumd/Sample.h"
#include "rumd/Timer.h"
#include "rumd/Device.h"
#include "rumd/IntegratorNVU.h"
#include "rumd/RUMD_Error.h"
#include "rumd/PairPotential.h"

int main( int argc, char* argv[]){
  try{
    int num_part = 1024;
    int parts_per_block = 16;
    int threads_per_part = 1;
    int num_io_blocks = 1;
    int io_block_size = 1;
    int write_trajectory = 1;
    int momentum_reset_interval = 100; // 0 to disable
    unsigned int quenchInterval = 50;
    int mode = 1; // 1 = run trajectory and write final info; 2 = run trajectory and and compare final info to previous runs.
    
    float rho=1.0;
    float dt = 1.f;
    float skin = 0.3f;
    float Rscal = 1.f;                   // Scaling of initial configuration
    float potentialEnergyPerParticle = 1.f;
    float displacementLength = 0.1f;
    //float4 sumU, sumK, sumW, sumS, sumC;
    //float stress[6];
    float totalDiffSq = 0.; // for mode 2

    while (1) {
      static struct option long_options[] = {
	{"n", 1, 0, 'n'},
	{"pb", 1, 0, 'p'},
	{"tp", 1, 0, 't'},
	{"thermostat", 1, 0, 'e'},
	{"numIOBlocks", 1, 0, 'o'},
	{"IOBlockSize", 1, 0, 'b'},
	{"writeTraj", 1, 0, 'w'},
	{"qi", 1, 0, 'q'},
	{"mode", 1, 0, 'm'},
	{"rho", 1, 0, 'l'},
	{"dt", 1, 0, 'd'},
	{"skin", 1, 0, 's'},
	{"Rscal", 1, 0, 'r'},
	{"dispLength", 1, 0, 'g'},
	{"U", 1, 0, 'a'},
	{0, 0, 0, 0}
      };
      int option_index = 0;
      int c = getopt_long(argc, argv, "",long_options, &option_index);
      if (c == -1)
	break;
      
      switch (c) {
      case 0:
	printf("option %s", long_options[option_index].name);
	if (optarg)
	  printf(" with arg %s", optarg);
	printf("\n");
	break;
      case 'n':
	num_part = atoi(optarg);
	break;
      case 'p':
	parts_per_block = atoi(optarg);
	break;
      case 't':
	threads_per_part = atoi(optarg);
	break;
      case 'o':
	num_io_blocks = atoi(optarg);
	break;
      case 'b':
	io_block_size = atoi(optarg);
	break; 
      case 'w':
	write_trajectory = atoi(optarg);
	break;
      case 'q':
	quenchInterval = (unsigned int)atoi(optarg);
	break;
      case 'm':
	mode = atoi(optarg);
	break;
      case 'l':
	rho = strtof(optarg, 0);
	break;
      case 'a':
	potentialEnergyPerParticle = strtof(optarg, 0);
	break;
      case 'g':
	displacementLength = strtof(optarg, 0);
	break;
      case 'd':
	dt = strtof(optarg, 0);
	break;
      case 's':
	skin = strtof(optarg, 0);
	break;
      case 'r':
	Rscal = strtof(optarg, 0);
	break;
      case '?':
	exit(EXIT_FAILURE);
	break;
      default:
	printf("?? getopt returned character code 0%o ??\n", c);
      }
    }
    
    Device::GetDevice().Init(argc, argv);

    // Create a new sample
    Sample* S = new Sample(parts_per_block, threads_per_part);
    
    // Create an integrator.
    IntegratorNVU* itg =  new IntegratorNVU();
    
    // Setup the integrator parameters.
    itg->SetTimeStep(dt);
    itg->SetDisplacementLength(displacementLength);
    itg->SetTargetPotentialEnergy(potentialEnergyPerParticle); 
    S->SetIntegrator(itg);
    
    // Create LJ pair-potential
    Pot_LJ_12_6_Gromacs potential;   
  
    // Setup parameters for Lennard-Jones interactions ( Kob & Andersen)
    //             i  j   Sig   Eps  Rcut (in units of Sigma_ij)
    potential.SetParams(0, 0, 1.00, 1.00, 2.50);
    potential.SetParams(0, 1, 0.80, 1.50, 2.50);
    potential.SetParams(1, 0, 0.80, 1.50, 2.50);
    potential.SetParams(1, 1, 0.88, 0.50, 2.50);    
    potential.SetNbListSkin(skin);
    S->SetPotential(&potential);
    
    potential.WritePotentials(S->GetSimulationBox());
    S->ReadConf("start.xyz.gz");

    itg->SetEuclideanConfiguration();        
    itg->SetMomentumToZero();
    unsigned numParticles = S->GetNumberOfParticles();

    // Calculate forces
    S->CalcF(); 
    //S->CalcEnergies(&sumU, &sumK, &sumW, &sumS, &sumC, stress);
    //float Tt = sumK.y;
    double pe_init = S->GetPotentialEnergy();
    double ke_init = S->GetIntegrator()->GetKineticEnergy();
    double vir_init = S->GetVirial();
    printf("Have initial conf; initial potential energy is %f\n",pe_init/numParticles);
    double startU_diff = -1.0, startK_diff = -1.0, startW_diff = -1.0, totalE=-1.0;
    FILE *start_energy_f;

    if(mode == 1) {
      start_energy_f = fopen("start_energies.dat","w");
      if(!start_energy_f){
	fprintf (stderr,"Error opening file start_energies.dat: %s\n", strerror(errno));
	exit(1);
      }
      fprintf(start_energy_f, "%.8f %.8f %.8f %.8f\n",  pe_init, ke_init, 
	    pe_init + ke_init, vir_init);
      fclose(start_energy_f);
    }
    else if (mode == 2){
      start_energy_f = fopen("start_energies.dat","r");
      if(!start_energy_f){
	fprintf (stderr,"Error opening file start_energies.dat: %s\n", strerror(errno));
	exit(1);
      }
      
      assert(start_energy_f);
      int f_stat = fscanf(start_energy_f, "%lf %lf %lf %lf", &startU_diff, &startK_diff, &totalE, &startW_diff);
      assert(f_stat == 4);
      startU_diff -= pe_init;
      startK_diff -= ke_init;
      startW_diff -= vir_init;
      startU_diff /= numParticles;
      startK_diff /= numParticles;
      startW_diff /= numParticles;
      totalDiffSq += startU_diff*startU_diff;
      totalDiffSq += startK_diff*startK_diff;
      totalDiffSq += startW_diff*startW_diff;
      fclose(start_energy_f);  
    }

    // Setup of various stuff
    S->SetOutputBlockSize(io_block_size);
    S->SetWriteTrajectory(write_trajectory);
    S->GetTrajectoryConfMetaData().has_pot_energies = true;
    S->GetTrajectoryConfMetaData().has_virials = true;
    S->GetTrajectoryConfMetaData().precision = 9;
    S->SetEnergyLinearWriteInterval(1);
    S->GetEnergiesMetaData().precision = 9;

    // Currently disabled in NVU.
    S->GetEnergiesMetaData().Set("totalEnergy",false);
    S->GetEnergiesMetaData().Set("temperature",false);
    S->GetEnergiesMetaData().Set("kineticEnergy",false);
    S->GetEnergiesMetaData().Set("pressure", false);
    S->GetEnergiesMetaData().Set("thermostat_Ps", false);
    
    // NVU specific output.
    S->GetEnergiesMetaData().Set("configTemperature", true);
    S->GetEnergiesMetaData().Set("laplacian", true); 
    S->GetEnergiesMetaData().Set("simulationDisplacementLength", true);
    S->GetEnergiesMetaData().Set("instantaneousTimeStepSq", true);
    S->GetEnergiesMetaData().Set("virial", false);
    S->GetEnergiesMetaData().Set("euclideanLength", true);
    
    Timer timer;
    // double eps = 1.e-10;

    unsigned long int num_iter = (unsigned long int) num_io_blocks * io_block_size + 1;
    
    printf("Doing %lu timesteps:\n", num_iter);
    for( unsigned long int i = 0; i < num_iter; i++) {
      // Calculate forces
      S->CalcF();
      S->NotifyOutputManagers(i);
      
      itg->Integrate(); 
      
      if( momentum_reset_interval > 0 && i % momentum_reset_interval==0 ) 
	itg->SetMomentumToZero();      
    }	
    
    double gpu_time = timer.elapsed();
    printf( "\nProcessing time GPU (s): %f\n", gpu_time);
    printf( " per timestep  GPU (ms): %f\n", 1000.0*gpu_time/num_iter);
    
    // Update forces and velocities
    S->CalcF();
    //S->CalcEnergies(&sumU, &sumK, &sumW, &sumS, &sumC, stress);
    //Tt = sumK.y;
    double pe_final = S->GetPotentialEnergy();
    double ke_final = S->GetIntegrator()->GetKineticEnergy();
    double vir_final = S->GetVirial();
    double endU_diff = 0., endK_diff = 0., endW_diff = 0.;
    if (mode == 0) {
      S->WriteConf("start.xyz");
    }
    if(mode == 1) {
      FILE *end_energy_f = fopen("end_energies.dat","w");
      if(!end_energy_f){
	fprintf (stderr,"Error opening file end_energies.dat: %s\n", strerror(errno));
	exit(1);
      }
      
      fprintf(end_energy_f, "%.8f %.8f %.8f %.8f %.8f\n", pe_final, ke_final, pe_final+ke_final, vir_final, gpu_time);
      fprintf(end_energy_f, "nSteps= %lu; dt=%f\n",num_iter, itg->GetTimeStep());
      fclose(end_energy_f);
      S->WriteConf("end.xyz");
    }
    else if(mode == 2) {
      printf("mode 2\n");
      FILE *end_energy_f = fopen("end_energies.dat","r");
      if(!end_energy_f){
	fprintf (stderr,"Error opening file end_energies.dat: %s\n", strerror(errno));
	exit(1);
      }
      
      double totalE, prevGpuTime;
      fscanf(end_energy_f, "%lf %lf %lf %lf %lf", &endU_diff, &endK_diff, &totalE, &endW_diff, &prevGpuTime);
      endU_diff -= pe_final;
      endK_diff -= ke_final;
      endW_diff -= vir_final;;
      endU_diff /= numParticles;
      endK_diff /= numParticles;
      endW_diff /= numParticles;
      
      totalDiffSq += endU_diff*endU_diff;
      totalDiffSq += endK_diff*endK_diff;
      totalDiffSq += endW_diff*endW_diff;
      FILE *statusFile = fopen("status.txt","w");
      if(!statusFile){
	fprintf (stderr,"Error opening file status.txt: %s\n", strerror(errno));
	exit(1);
      }
      	  
      fprintf(statusFile, "Status for N=%d, rho=%.2f, U=%.5f, pb=%d, tp=%d, ",numParticles,rho,potentialEnergyPerParticle,parts_per_block,threads_per_part);
      fprintf(statusFile, "NVU : ");
      if(totalDiffSq < 1.e-6)
      // need to allow for changes in round-off error when different
      // (pb, tp) combinations are used so tolerance should not be too small
	fprintf(statusFile, "OK\n");
      else
	fprintf(statusFile, "FAIL\n");
      
      fprintf(statusFile,"Total difference squared: %g\n",totalDiffSq);
      fprintf(statusFile,"startU_diff: %.8f\n",startU_diff);
      fprintf(statusFile,"startK_diff: %.8f\n",startK_diff);
      fprintf(statusFile,"startW_diff: %.8f\n",startW_diff);
      fprintf(statusFile,"endU_diff: %.8f\n",endU_diff);
      fprintf(statusFile,"endK_diff: %.8f\n",endK_diff);
      fprintf(statusFile,"endW_diff: %.8f\n",endW_diff);
      fprintf(statusFile,"time taken, before/now (s): %f/%f\n",prevGpuTime, gpu_time);
    }
    delete S;
  } catch(RUMD_Error e) {
    std::cerr << "RUMD_Error thrown from function " << e.className << "::" << e.methodName << ": " << std::endl << e.errorStr;
    return 1;
  }
  return 0;
}
