/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include <getopt.h>
#include <iostream>

#include "rumd/Timer.h"
#include "rumd/Device.h"
#include "rumd/Sample.h"
#include "rumd/RUMD_Error.h"
#include "rumd/IntegratorNVT.h"
#include "rumd/IntegratorIHS.h"
#include "rumd/PairPotential.h"

int main( int argc, char* argv[] ){
  try{
    int parts_per_block = 16;
    int threads_per_part = 8;
    int thermostat = 0;                  // 0=none 1=Nose-Hoover
    int num_io_blocks = 1;               // Number of (logarithmic) configuration blocks
    int io_block_size = 1;
    int restartBlock = -1;               // which block to restart from (-1 means no restart)
    int dump_energies = 200;             // Interval (timesteps) between dumping of energies (0 means do not write energies)
    int dump_ihs = 1;                    // Quench every dump_ihs time steps.
    int momentum_reset_interval = 100;   // 0 to disable
    int ihs_iter = 1;                    // Number of iterations to do a quench.
    
    float Ti = 1.f;                      // Initial temperature
    float Tf = 1.f;                      // Final temperature
    float dt = 0.0025f;
    float dt_ihs = 0.01f;
    float skin = 0.3f;
    float Rscal = 1.f;                   // Scaling of initial configuration
    float relaxation = 0.3f;
    
    bool Tf_assigned = false;
    
    while (1) {
      static struct option long_options[] = {
	{"pb", 1, 0, 'p'},
	{"tp", 1, 0, 't'},
	{"thermostat", 1, 0, 'e'},
	{"numIOBlocks", 1, 0, 'o'},
	{"IOBlockSize", 1, 0, 'b'},
	{"dump_nrg", 1, 0, 'w'},
	{"dump_ihs", 1, 0, 'a'},
	{"restartBlock", 1, 0, 'x'},
	{"Ti", 1, 0, 'i'},
	{"Tf", 1, 0, 'f'},
	{"dt", 1, 0, 'd'},
	{"relaxation", 1, 0, 'u'},
	{"skin", 1, 0, 's'},
	{"Rscal", 1, 0, 'r'},
	{"ihs_iter", 1, 0, 'c'},
	{0, 0, 0, 0}
      };
      int option_index = 0;
      int c = getopt_long(argc, argv, "",long_options, &option_index);
      if (c == -1)
	break;
      
      switch (c) {
      case 0:
	printf("option %s", long_options[option_index].name);
	if (optarg)
	  printf(" with arg %s", optarg);
        printf("\n");
	break;
      case 'p':
	parts_per_block = atoi(optarg);
	break;
      case 't':
	threads_per_part = atoi(optarg);
	break;
      case 'e':
	thermostat = atoi(optarg);
	break;
      case 'o':
	num_io_blocks = atoi(optarg);
	break;
      case 'b':
      io_block_size = atoi(optarg);
      break; 
      case 'w':
	dump_energies = atoi(optarg);
	break;
      case 'a':
	dump_ihs = atoi(optarg);
	break;
      case 'x':
	restartBlock = atoi(optarg);
	break;
      case 'i':
	Ti = strtof(optarg, 0);
	break;
      case 'f':
	Tf = strtof(optarg, 0);
	Tf_assigned = true;
      break;
      case 'd':
	dt = strtof(optarg, 0);
	break;
      case 'u':
	relaxation = strtof(optarg, 0);
      break;
      case 's':
	skin = strtof(optarg, 0);
	break;
      case 'r':
	Rscal = strtof(optarg, 0);
	break;
      case 'c':
	ihs_iter = strtof(optarg, 0);
	break;
      case '?':
	exit(EXIT_FAILURE);
	break;
      default:
	printf("?? getopt returned character code 0%o ??\n", c);
      }
    }
    
    if (!Tf_assigned) Tf = Ti;
    
    Device::GetDevice().Init(argc, argv);
    
    // Create samples
    Sample* S = new Sample(parts_per_block, threads_per_part);
    Sample* SI = new Sample(parts_per_block, threads_per_part);
    
    // Create integrators.
    IntegratorNVT* itg =  new IntegratorNVT(dt, Ti);
    IntegratorIHS* itgI =  new IntegratorIHS(dt_ihs);
    
    // Setup the integrator parameters.
    itg->SetRelaxationTime(relaxation);
    itg->SetThermostatOn(thermostat);
    S->SetIntegrator(itg);
    SI->SetIntegrator(itgI);

    // Create the box geometry.
    //RectangularSimulationBox* simBox = new RectangularSimulationBox();
    //S->SetSimulationBox(simBox);
    //SI->SetSimulationBox(simBox);

    std::cout << "NVT pair potential" << std::endl;
    Pot_LJ_12_6_Gromacs potential;   
    potential.SetParams(0, 0, 1.00, 1.00, 2.5);
    potential.SetParams(0, 1, 0.80, 1.50, 2.5);
    potential.SetParams(1, 0, 0.80, 1.50, 2.5);
    potential.SetParams(1, 1, 0.88, 0.50, 2.5);
    potential.SetNbListSkin(skin);
    S->SetPotential(&potential);
    
    std::cout << "IHS pair potential" << std::endl;
    Pot_LJ_12_6_Gromacs potentialI;   
    potentialI.SetParams(0, 0, 1.00, 1.00, 2.5);
    potentialI.SetParams(0, 1, 0.80, 1.50, 2.5);
    potentialI.SetParams(1, 0, 0.80, 1.50, 2.5);
    potentialI.SetParams(1, 1, 0.88, 0.50, 2.5);
    potentialI.SetNbListSkin(skin);
    SI->SetPotential(&potentialI);
    
    // Read configuration & Copy it to device
    unsigned long int initialTimeStep = 0;
    
    if(restartBlock < 0) {
      S->ReadConf("start.xyz.gz");
      SI->ReadConf("start.xyz.gz");     
      S->IsotropicScaleSystem(Rscal);
      SI->IsotropicScaleSystem(Rscal);
      itg->SetMomentumToZero();
    }  
    else{
      S->ReadRestartConf(restartBlock);
      SI->ReadRestartConf(restartBlock);
      initialTimeStep = (unsigned long int) (restartBlock)*io_block_size;
    }

    potential.WritePotentials(S->GetSimulationBox());
    //S->DumpPotentials();
    
    // Perform a single quench.
    *SI = *S;
    for( int j=0; j < ihs_iter; j++ ){
      SI->CalcF();
      itgI->Integrate();
    }
    
    // Set the initial config for IHS comparison.
    itgI->SetPreviousInherentStateConfiguration();
    
    // Setup of various stuff
    S->SetOutputBlockSize(io_block_size);
    S->SetWriteTrajectory(true);
    if(dump_energies) {
      S->SetWriteEnergies(true);
      S->SetEnergyLinearWriteInterval(dump_energies);
    }
    
    S->GetTrajectoryConfMetaData().has_pot_energies = true;
    S->GetTrajectoryConfMetaData().has_virials = true;
    S->GetTrajectoryConfMetaData().precision = 9;
    S->GetEnergiesMetaData().precision = 9;
    
    if(num_io_blocks > 9999){
      fprintf(stderr, "Error: Number of IO blocks cannot be more than 9999.\n");
      exit(1);
    }
    
    unsigned long int num_iter = (unsigned long int) num_io_blocks*io_block_size + 1;
    
    printf("Total number of quench iteration(s) is %i, quenching configurations every %d timestep.\n", ihs_iter, dump_ihs);
    printf("Total number of timestep(s) is %lu, ", num_iter);
    if(restartBlock >= 0)
      printf("restart: starting from iteration %lu\n",initialTimeStep);
    printf("dumping energies every %d timestep (every %f time-unit).\n", dump_energies, dump_energies*itg->GetTimeStep());
    
    // Execute the kernel(s)
    Timer timer;
    
    FILE* IHSData; IHSData = fopen("ihs_sampled.dat", "w");
    
    for( unsigned long int i = initialTimeStep; i < num_iter; i++ ){
      S->CalcF(); // Calculate forces
      S->NotifyOutputManagers(i); // Called here to ensure synchronization
      
      // Copies the minimum GPU sample for integration to continue.
      if( dump_ihs > 0 && i % dump_ihs == 0 ){
	*SI = *S; 
	
	for( int j=0; j < ihs_iter; j++ ){
	  SI->CalcF(); 
	  itgI->Integrate();
	}
	
	// Detect a transition between inherent states.
	float length = itgI->GetInherentStateTransitionLength();
	//if( length > 0.1f ){
	// itgI->DumpInherentStateTransitionConfigurations();      
	// itgI->SetPreviousInherentStateConfiguration();
	//}      
	
	fprintf(IHSData, "%e %e %e\n", itgI->GetInherentStatePotentialEnergy(), itgI->GetInherentStateForceSquared(), length);
      } 
      
      itg->Integrate(); // Integrate NVT dynamics.
      
      if ( momentum_reset_interval > 0 && i % momentum_reset_interval==0 ) 
	itg->SetMomentumToZero();
    } 
    
    double gpu_time = timer.elapsed();
    printf( "\nProcessing time GPU (s): %f\n", gpu_time);
    printf( " N  thermostat  pb  tp  ms/timestep: \n");
    float t_per_step = 0.;
    
    if (num_iter > 0) t_per_step = 1000.0*gpu_time/num_iter;
    printf( "%d %d   %d   %d  %f\n", S->GetNumberOfParticles(), thermostat, parts_per_block, threads_per_part, t_per_step);
    
    // Update forces and velocities and copy everything back from device 
    S->CalcF();  
    S->WriteConf("confout.xyz");
    
    // Free objects.
    delete S; delete itg; delete itgI;
    //delete simBox;
  } 
  catch(RUMD_Error e){
    std::cerr << "RUMD_Error thrown from function " << e.className << "::" << e.methodName << ": " << std::endl << e.errorStr << std::endl;
    return 1;
  }
  return 0;  
}
