
/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/ConstraintPotential.h"
#include "rumd/MoleculeData.h"
#include "rumd/Sample.h"
#include "rumd/rumd_algorithms.h"

#include <iostream>

using namespace std;

template <class S> __global__ void test( float* A, float* b, float4* position, float4* velocity, float4* force, S* simBox, float* simBoxPointer );
template <class S> __global__ void test1( float* x, float4* position, float4* force, S* simBox, float* simBoxPointer );

//////////////////////////////////////////////////
// Constructors 
//////////////////////////////////////////////////

ConstraintPotential::ConstraintPotential( MoleculeData* M, Sample* S ){
  this->S = S; 
  this->M = M;
  
  numberOfMolecules = M->GetNumberOfMolecules();
  
  if( cudaMalloc( (void**) &d_dimLinearSystems, numberOfMolecules * sizeof(unsigned) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("ConstraintPotential","ConstraintPotential","Malloc failed on d_dimLinearSystems") );
  
  unsigned* constraintList = (unsigned*) malloc( numberOfMolecules * sizeof(unsigned) );
  
  unsigned totalNumberOfConstraints = 0;
  unsigned totalSquaredNumberOfConstraints = 0;
  
  for(unsigned i = 0; i < numberOfMolecules; i++){
    // Do something more generic from MoleculeData info.
    constraintList[i] = 1;
    totalNumberOfConstraints += constraintList[i];
    totalSquaredNumberOfConstraints += ( constraintList[i] * constraintList[i] );
  }
  
  // Copy the dimensions of the constraint systems.
  cudaMemcpy( d_dimLinearSystems, constraintList, numberOfMolecules * sizeof(unsigned), cudaMemcpyHostToDevice );
  
  if( cudaMalloc( (void**) &d_A, totalSquaredNumberOfConstraints * sizeof(float) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("ConstraintPotential","ConstraintPotential","Malloc failed on d_A") );
  
  if( cudaMalloc( (void**) &d_b, totalNumberOfConstraints * sizeof(float) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("ConstraintPotential","ConstraintPotential","Malloc failed on d_b") );

  free(constraintList);
}

ConstraintPotential::~ConstraintPotential(){
  cudaFree(d_dimLinearSystems);
  cudaFree(d_A);
  cudaFree(d_b);
}

//////////////////////////////////////////////////
// Class methods 
//////////////////////////////////////////////////

void ConstraintPotential::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs){
  BuildLinearConstraintSystem();
  solveLinearSystems( d_b, d_A, d_dimLinearSystems, numberOfMolecules );  
  UpdateTotalForce();
  WritePotential();
}

void ConstraintPotential::BuildLinearConstraintSystem(){
  RectangularSimulationBox* testRSB = dynamic_cast<RectangularSimulationBox*>(S->GetSimulationBox());
  test<<< numberOfMolecules, 1 >>>( d_A, d_b, S->particleData.d_r, S->particleData.d_v, S->particleData.d_f, 
				    testRSB, testRSB->GetDevicePointer() );
}

void ConstraintPotential::UpdateTotalForce(){ 
  RectangularSimulationBox* testRSB = dynamic_cast<RectangularSimulationBox*>(S->GetSimulationBox());
  test1<<< numberOfMolecules, 1 >>>( d_b, S->particleData.d_r, S->particleData.d_f, testRSB, testRSB->GetDevicePointer() );
}

void ConstraintPotential::WritePotential(){
  double total_diff = 0, rms = 0;
  S->particleData.CopyPosFromDevice();
  
  // Read constraints.
  for( unsigned k = 0; k < 500; k++){
    int first_p =  2 * k;
    int second_p = 2 * k + 1;
    float distance = 0.584324;
    
    float4 r_1 = S->particleData.h_r[first_p];
    float4 r_2 = S->particleData.h_r[second_p];
    
    RectangularSimulationBox* testRSB = dynamic_cast<RectangularSimulationBox*>(S->GetSimulationBox());
    
    float4 r_sq = testRSB->calculateDistance( r_1, r_2, testRSB->GetHostPointer() );
    double r = sqrt( r_sq.w );
    double difference = r - distance;
    
    total_diff += difference;
    rms += difference * difference;
  }
  
  float2 deviations = { total_diff / double(500), sqrt( rms / double(500) ) };
  cout << "=== Constraints: ( " << deviations.x << " , " << deviations.y << " ) === " << endl;
}

//////////////////////////////////////////////////
// Build Linear system
//////////////////////////////////////////////////

template < class S > __global__ void test( float* A, float* b, float4* position, float4* velocity, float4* force,  S* simBox, float* simBoxPointer ){
  // Load the simulation box in local memory to avoid bank conflicts.
  float array[simulationBoxSize];
  simBox->loadMemory( array, simBoxPointer, 1 );

  float4 my_position1 = position[2 * MyGP];
  float4 my_position2 = position[2 * MyGP + 1];
  
  float4 my_velocity1 = velocity[2 * MyGP];
  float4 my_velocity2 = velocity[2 * MyGP + 1];
  
  float4 my_force1 = force[2 * MyGP];
  float4 my_force2 = force[2 * MyGP + 1];
  
  float timeStep = 0.0025f;
  
  float4 before = simBox->calculateDistance( my_position1, my_position2, array );

  // Update to v(t+h/2).
  my_velocity1.x += my_velocity1.w * my_force1.x * timeStep; 
  my_velocity1.y += my_velocity1.w * my_force1.y * timeStep;
  my_velocity1.z += my_velocity1.w * my_force1.z * timeStep;
  
  // Update to r(t+h).
  my_position1.x += my_velocity1.x * timeStep; 
  my_position1.y += my_velocity1.y * timeStep; 
  my_position1.z += my_velocity1.z * timeStep; 

  // Update to v(t+h/2).
  my_velocity2.x += my_velocity2.w * my_force2.x * timeStep; 
  my_velocity2.y += my_velocity2.w * my_force2.y * timeStep;
  my_velocity2.z += my_velocity2.w * my_force2.z * timeStep;
  
  // Update to r(t+h).
  my_position2.x += my_velocity2.x * timeStep; 
  my_position2.y += my_velocity2.y * timeStep; 
  my_position2.z += my_velocity2.z * timeStep; 

  float4 local_diff = simBox->calculateDistance( my_position1, my_position2, array );

  A[MyGP] = - 2.f * (1.f / 0.163173827f) * timeStep *  timeStep * ( local_diff.x * before.x + local_diff.y * before.y + local_diff.z * before.z );
  b[MyGP] = local_diff.w - ( 0.584324f * 0.584324f );
}

//////////////////////////////////////////////////
// Update force with constraint force
//////////////////////////////////////////////////

template < class S > __global__ void test1( float* x, float4* position, float4* force, S* simBox, float* simBoxPointer ){
  // Load the simulation box in local memory to avoid bank conflicts.
  float array[simulationBoxSize];
  simBox->loadMemory( array, simBoxPointer, 1 );

  //printf("%u <--> %lf\n",MyGP, x[MyGP]);
  
  float4 my_position1 = position[2 * MyGP];
  float4 my_position2 = position[2 * MyGP + 1];

  float4 my_force1 = force[2 * MyGP];
  float4 my_force2 = force[2 * MyGP + 1];
  
  float4 before = simBox->calculateDistance( my_position1, my_position2, array );

  my_force1.x += x[MyGP] * before.x;
  my_force1.y += x[MyGP] * before.y;
  my_force1.z += x[MyGP] * before.z;

  my_force2.x -= x[MyGP] * before.x;
  my_force2.y -= x[MyGP] * before.y;
  my_force2.z -= x[MyGP] * before.z;
  
  force[2 * MyGP] = my_force1;
  force[2 * MyGP + 1] = my_force2;
}
